package com.sun.xml.internal.txw2;

import java.util.Iterator;
import java.util.HashMap;
import java.util.Map;
import com.sun.xml.internal.txw2.output.XmlSerializer;

public final class Document
{
    private final XmlSerializer out;
    private boolean started;
    private Content current;
    private final Map<Class, DatatypeWriter> datatypeWriters;
    private int iota;
    private final NamespaceSupport inscopeNamespace;
    private NamespaceDecl activeNamespaces;
    private final ContentVisitor visitor;
    private final StringBuilder prefixSeed;
    private int prefixIota;
    static final char MAGIC = '\0';
    
    Document(XmlSerializer out) {
        this.started = false;
        this.current = null;
        this.datatypeWriters = new HashMap();
        this.iota = 1;
        this.inscopeNamespace = new NamespaceSupport();
        this.visitor = new ContentVisitor() {
            @Override
            public void onStartDocument() {
                throw new IllegalStateException();
            }
            
            @Override
            public void onEndDocument() {
                Document.this.out.endDocument();
            }
            
            @Override
            public void onEndTag() {
                Document.this.out.endTag();
                Document.this.inscopeNamespace.popContext();
                Document.this.activeNamespaces = null;
            }
            
            @Override
            public void onPcdata(StringBuilder buffer) {
                if (Document.this.activeNamespaces != null) {
                    buffer = Document.access$300(Document.this, buffer);
                }
                Document.this.out.text(buffer);
            }
            
            @Override
            public void onCdata(StringBuilder buffer) {
                if (Document.this.activeNamespaces != null) {
                    buffer = Document.access$300(Document.this, buffer);
                }
                Document.this.out.cdata(buffer);
            }
            
            @Override
            public void onComment(StringBuilder buffer) {
                if (Document.this.activeNamespaces != null) {
                    buffer = Document.access$300(Document.this, buffer);
                }
                Document.this.out.comment(buffer);
            }
            
            @Override
            public void onStartTag(String nsUri, String localName, Attribute attributes, NamespaceDecl namespaces) {
                assert nsUri != null;
                assert localName != null;
                Document.this.activeNamespaces = namespaces;
                if (!Document.this.started) {
                    Document.this.started = true;
                    Document.this.out.startDocument();
                }
                Document.this.inscopeNamespace.pushContext();
                for (NamespaceDecl ns = namespaces; ns != null; ns = ns.next) {
                    ns.declared = false;
                    if (ns.prefix != null) {
                        String uri = Document.this.inscopeNamespace.getURI(ns.prefix);
                        if (uri == null || !uri.equals(ns.uri)) {
                            Document.this.inscopeNamespace.declarePrefix(ns.prefix, ns.uri);
                            ns.declared = true;
                        }
                    }
                }
                for (NamespaceDecl ns = namespaces; ns != null; ns = ns.next) {
                    if (ns.prefix == null) {
                        if (Document.this.inscopeNamespace.getURI("").equals(ns.uri)) {
                            ns.prefix = "";
                        }
                        else {
                            String p = Document.this.inscopeNamespace.getPrefix(ns.uri);
                            if (p == null) {
                                while (Document.this.inscopeNamespace.getURI(p = Document.access$500(Document.this)) != null) {}
                                ns.declared = true;
                                Document.this.inscopeNamespace.declarePrefix(p, ns.uri);
                            }
                            ns.prefix = p;
                        }
                    }
                }
                assert !(!namespaces.uri.equals(nsUri));
                assert namespaces.prefix != null : "a prefix must have been all allocated";
                Document.this.out.beginStartTag(nsUri, localName, namespaces.prefix);
                for (NamespaceDecl ns = namespaces; ns != null; ns = ns.next) {
                    if (ns.declared) {
                        Document.this.out.writeXmlns(ns.prefix, ns.uri);
                    }
                }
                for (Attribute a = attributes; a != null; a = a.next) {
                    String prefix;
                    if (a.nsUri.length() == 0) {
                        prefix = "";
                    }
                    else {
                        prefix = Document.this.inscopeNamespace.getPrefix(a.nsUri);
                    }
                    Document.this.out.writeAttribute(a.nsUri, a.localName, prefix, Document.access$300(Document.this, a.value));
                }
                Document.this.out.endStartTag(nsUri, localName, namespaces.prefix);
                return;
            }
        };
        this.prefixSeed = new StringBuilder("ns");
        this.prefixIota = 0;
        this.out = out;
        for (DatatypeWriter dw : DatatypeWriter.BUILTIN) {
            this.datatypeWriters.put(dw.getType(), dw);
        }
    }
    
    void flush() {
        this.out.flush();
    }
    
    void setFirstContent(Content c) {
        assert this.current == null;
        (this.current = new StartDocument()).setNext(this, c);
        return;
    }
    
    public void addDatatypeWriter(DatatypeWriter<?> dw) {
        this.datatypeWriters.put(dw.getType(), dw);
    }
    
    void run() {
        while (true) {
            Content next = this.current.getNext();
            if (next == null || !next.isReadyToCommit()) {
                break;
            }
            next.accept(this.visitor);
            next.written();
            this.current = next;
        }
    }
    
    void writeValue(Object obj, NamespaceResolver nsResolver, StringBuilder buf) {
        if (obj == null) {
            throw new IllegalArgumentException("argument contains null");
        }
        if ((obj instanceof Object[])) {
            for (Object o2 : (Object[])(Object[])obj) {
                this.writeValue(o2, nsResolver, buf);
            }
            return;
        }
        if ((obj instanceof Iterable)) {
            for (Object o : (Iterable)obj) {
                this.writeValue(o, nsResolver, buf);
            }
            return;
        }
        if (buf.length() > 0) {
            buf.append(' ');
        }
        Class<Object> c = obj.getClass();
        while (c != null) {
            DatatypeWriter<Object> dw = (DatatypeWriter)this.datatypeWriters.get(c);
            if (dw != null) {
                dw.print(obj, nsResolver, buf);
                return;
            }
            c = c.getSuperclass();
        }
        buf.append(obj);
        return;
    }
    
    private String newPrefix() {
        this.prefixSeed.setLength(2);
        this.prefixSeed.append(++this.prefixIota);
        return this.prefixSeed.toString();
    }
    
    private StringBuilder fixPrefix(StringBuilder buf) {
        assert this.activeNamespaces != null;
        int len;
        int i;
        for (len = buf.length(), i = 0; i < len && buf.charAt(i) != '\0'; ++i) {}
        if (i == len) {
            return buf;
        }
        while (i < len) {
            char uriIdx;
            NamespaceDecl ns;
            for (uriIdx = buf.charAt(i + 1), ns = this.activeNamespaces; ns != null && ns.uniqueId != uriIdx; ns = ns.next) {}
            if (ns == null) {
                throw new IllegalStateException("Unexpected use of prefixes " + buf);
            }
            int length = 2;
            String prefix = ns.prefix;
            if (prefix.length() == 0) {
                if (buf.length() <= i + 2 || buf.charAt(i + 2) != ':') {
                    throw new IllegalStateException("Unexpected use of prefixes " + buf);
                }
                length = 3;
            }
            buf.replace(i, i + length, prefix);
            for (len += prefix.length() - length; i < len && buf.charAt(i) != '\0'; ++i) {}
        }
        return buf;
    }
    
    char assignNewId() {
        return (char)(this.iota++);
    }
    
    static StringBuilder access$300(Document x0, StringBuilder x1) {
        return x0.fixPrefix(x1);
    }
    
    static String access$500(Document x0) {
        return x0.newPrefix();
    }
}
